@@ -1,3 +1,7 @@
+"""
+Helper functions for SIESTA runs or analysis of SIESTA log files
+"""
+from __future__ import absolute_import
 import os
 import re
 import shutil
@@ -14,9 +18,13 @@ def get_it(files):
 def read_fdf(fdfpath, geo):
     """Read FDF file"""
     print(f"Reading {fdfpath}")
-    with open(fdfpath, "r") as fdffile:
+    with open(fdfpath, "r", encoding="utf-8") as fdffile:
         fdf = fdffile.read()
-        ind = fdf.split("%block ChemicalSpeciesLabel\n")[1].split("%endblock ChemicalSpeciesLabel\n")[0]
+        ind = fdf.split(
+            "%block ChemicalSpeciesLabel\n"
+        )[1].split(
+            "%endblock ChemicalSpeciesLabel\n"
+        )[0]
         ind = ind.splitlines()
         for i in ind:
             for g in geo:
@@ -30,7 +38,7 @@ def read_fdf(fdfpath, geo):
 def create_fdf(fdf, geo, newfdfpath, number):
     """Create new FDF file"""
     print(f"Creating {newfdfpath}")
-    with open(newfdfpath, "w") as newfdffile:
+    with open(newfdfpath, "w", encoding="utf-8") as newfdffile:
         newfdf = fdf.split("%block AtomicCoordinatesAndAtomicSpecies\n")[0]
         newfdf += "%block AtomicCoordinatesAndAtomicSpecies\n"
         for g in geo:
@@ -52,7 +60,7 @@ def read_energy(energies=[], files=None, it=[], print_=True):
     for f in files:
         if print_:
             print(f)
-        with open(f, "r") as file:
+        with open(f, "r", encoding="utf-8") as file:
             lines = file.readlines()
             for line in lines:
                 if line.startswith("siesta:         Total =  "):
@@ -66,70 +74,70 @@ def print_run(for_, cores, conda):
     print(
         f"""Running SIESTA for {for_}
                 {f' in parallel with {cores} cores' if cores is not None else ''}
-                {' in conda' if conda else ''}"""
+                {' in conda' if conda else ''}""".replace("\n", "").replace("                ", "")
     )
 
 
-def check_restart(fdffile, i, label, cwd, cont, contextensions):
+def check_restart(*, fdffile, i, label, cwd, cont, contextensions):
     """Check DM, XV, CG, and LWF parameters in an FDF file"""
     fdf = fdffile.read()
     if "DM" in contextensions:
         check_restart_ext(
-            "DM",
-            fdf,
-            r"# *DM\.UseSaveDM +\.true\.",
-            r"DM\.UseSaveDM +\.false\.",
-            "DM.UseSaveDM        .true.",
-            "DM.UseSaveDM",
-            cwd,
-            i,
-            cont,
-            label
+            ext="DM",
+            fdf=fdf,
+            match1=r"# *DM\.UseSaveDM +(\.true\.|T)",
+            match2=r"DM\.UseSaveDM +(\.false\.|F)",
+            repl="DM.UseSaveDM        .true.",
+            out="DM.UseSaveDM",
+            cwd=cwd,
+            i=i,
+            cont=cont,
+            label=label
         )
     if "XV" in contextensions:
         check_restart_ext(
-            "XV",
-            fdf,
-            r"# *MD\.UseSaveXV +\.true\.",
-            r"MD\.UseSaveXV +\.false\.",
-            "MD.UseSaveXV        .true.",
-            "MD.UseSaveCG",
-            cwd,
-            i,
-            cont,
-            label
+            ext="XV",
+            fdf=fdf,
+            match1=r"# *MD\.UseSaveXV +(\.true\.|T)",
+            match2=r"MD\.UseSaveXV +(\.false\.|F)",
+            repl="MD.UseSaveXV        .true.",
+            out="MD.UseSaveCG",
+            cwd=cwd,
+            i=i,
+            cont=cont,
+            label=label
         )
     if "CG" in contextensions:
         check_restart_ext(
-            "CG",
-            fdf,
-            r"# *MD\.UseSaveCG +\.true\.",
-            r"MD\.UseSaveCG +\.false\.",
-            "MD.UseSaveCG        .true.",
-            "MD.UseSaveCG",
-            cwd,
-            i,
-            cont,
-            label
+            ext="CG",
+            fdf=fdf,
+            match1=r"# *MD\.UseSaveCG +(\.true\.|T)",
+            match2=r"MD\.UseSaveCG +(\.false\.|F)",
+            repl="MD.UseSaveCG        .true.",
+            out="MD.UseSaveCG",
+            cwd=cwd,
+            i=i,
+            cont=cont,
+            label=label
         )
     if "LWF" in contextensions:
         check_restart_ext(
-            "LWF",
-            fdf,
-            r"# *ON\.UseSaveLWF +\.true\.",
-            r"ON\.UseSaveLWF +\.false\.",
-            "ON.UseSaveLWF        .true.",
-            "ON.UseSaveLWF",
-            cwd,
-            i,
-            cont,
-            label
+            ext="LWF",
+            fdf=fdf,
+            match1=r"# *ON\.UseSaveLWF +(\.true\.|T)",
+            match2=r"ON\.UseSaveLWF +(\.false\.|F)",
+            repl="ON.UseSaveLWF        .true.",
+            out="ON.UseSaveLWF",
+            cwd=cwd,
+            i=i,
+            cont=cont,
+            label=label
         )
     fdffile.seek(0)
     fdffile.write(fdf)
 
 
-def check_restart_ext(ext, fdf, match1, match2, repl, out, cwd, i, cont, label):
+def check_restart_ext(*, ext, fdf, match1, match2, repl, out, cwd, i, cont, label):
     """Check DM, XV, CG, and LWF parameters in an FDF file individually"""
     match = re.search(match1, fdf)
     if match is None:
@@ -140,16 +148,19 @@ def check_restart_ext(ext, fdf, match1, match2, repl, out, cwd, i, cont, label):
     else:
         print(f"Setting '{out}' as '.true.' in {cwd}{os.sep}i{i}{os.sep}{cont}{os.sep}{label}.fdf")
         fdf = fdf.replace(match[0], repl)
-    if ext == "DM" and (re.search("WriteDM +.true.", fdf) is None or re.search("# *WriteDM +.true.", fdf) is not None
+    if ext == "DM" and (re.search("WriteDM +.true.", fdf) is None
+                        or re.search("# *WriteDM +.true.", fdf) is not None
                         or re.search("WriteDM +.false.", fdf) is not None):
         print(
-            f"WARNING: 'WriteDM             .true.' not found in {cwd}{os.sep}i{i}{os.sep}{cont}{os.sep}{label}.fdf"
+            f"WARNING: 'WriteDM .true.' not found in {cwd}{os.sep}i{i}{os.sep}{cont}" +
+            f"{os.sep}{label}.fdf"
         )
 
 
 def check_userbasis(fdffile):
-    with open(fdffile, "r") as f:
-        if re.search(r"Userbasis *\.true\.", f.read()):
+    """Check if the Userbasis parameter in the fdf file is either true or false"""
+    with open(fdffile, "r", encoding="utf-8") as f:
+        if re.search(r"Userbasis *(\.true\.|T)", f.read()):
             return True
         f.close()
         return False
@@ -161,7 +172,7 @@ def copy_file(sourcefile, destinationfile):
         raise FileNotFoundError(f"ERROR: {sourcefile} is not found")
     try:
         print(f"Copying {sourcefile} to {destinationfile}")
-        if not os.path.isfile(destinationfile):
+        if not os.path.exists(destinationfile):
             shutil.copy(sourcefile, destinationfile)
             print(f"{sourcefile} is copied to {destinationfile} successfully")
         else:
@@ -170,8 +181,8 @@ def copy_file(sourcefile, destinationfile):
         print(f"ERROR: {sourcefile} and {destinationfile} represents the same file")
     except PermissionError:
         print(f"ERROR: Permission denied while copying {sourcefile} to {destinationfile}")
-    except Exception:
-        print(f"ERROR: An error occurred while copying {sourcefile} to {destinationfile}")
+    except (shutil.Error, OSError, IOError) as e:
+        print(f"ERROR: An error occurred while copying {sourcefile} to {destinationfile} ({e})")
 
 
 def sort_(files, path, cont):
@@ -181,9 +192,10 @@ def sort_(files, path, cont):
     match = [re.search(f"{path}({os.sep}{cont}_*([0-9]*))*", f) for f in files]
     sortedmatch = [[m[0], m[1], m[2], m[3]] for m in match]
     sortedmatch = [x for _, x in sorted(zip(
-        [int(f"{m[1]}0") if m[3] is None else int(f"{m[1]}1") if m[3] == "" else int(m[1] + m[3]) for m in
-         sortedmatch
-         ], sortedmatch
+        [int(f"{m[1]}0") if m[3] is None else
+         int(f"{m[1]}1") if m[3] == "" else
+         int(m[1] + m[3]) for m in sortedmatch
+        ], sortedmatch
     ))]
     for s in sortedmatch:
         for f in files:
